#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.

from threading import Lock


class SeqNo:
    def __init__(self):
        self.content_related_messages_sent = 0
        self.lock = Lock()

    def __call__(self, is_content_related: bool) -> int:
        with self.lock:
            seq_no = (self.content_related_messages_sent * 2) + (1 if is_content_related else 0)

            if is_content_related:
                self.content_related_messages_sent += 1

            return seq_no
